import pyglet
import pyglet.gl as gl
import pyglet.graphics as graphics
import numpy as np
from typing import NamedTuple
import os

class simulator_world(NamedTuple):
     lanes:list = []
     cars:list = []
     objects:list = []

class simulator:
    def __init__(self, dt =0.1, fullscreen=False, name='Simulator', magnify=1., iters=100, sprite_scale= 0.15/600.,
                 window_size= 600):
        self.dt = dt
        self.fullscreen = fullscreen
        self.name = name
        self.magnify = magnify
        self.max_iters = iters
        self.iter = 0
        self.window_size = window_size

        working_dir = os.path.dirname(os.path.realpath(__file__))
        pyglet.resource.path = [os.path.join(working_dir, 'imgs')]
        pyglet.resource.reindex()

        def centered_image(filename):
            img = pyglet.resource.image(filename)
            img.anchor_x = img.width / 2.
            img.anchor_y = img.height / 2.
            return img

        def car_sprite(color, scale=sprite_scale):
            sprite = pyglet.sprite.Sprite(centered_image('car-{}.png'.format(color)), subpixel=True)
            sprite.scale = scale
            return sprite

        def object_sprite(name, scale=sprite_scale):
            sprite = pyglet.sprite.Sprite(centered_image('{}.png'.format(name)), subpixel=True)
            sprite.scale = scale
            return sprite

        self.sprites = {c: car_sprite(c) for c in ['red', 'yellow', 'purple', 'white', 'orange', 'gray', 'blue']}
        self.obj_sprites = {c: object_sprite(c) for c in ['cone', 'firetruck']}

    def init_simulator(self, world, task_name):
        self.event_loop = pyglet.app.EventLoop()
        self.window = pyglet.window.Window(self.window_size, self.window_size,
                                           fullscreen=self.fullscreen, caption=self.name)
        self.background = pyglet.resource.texture('grass.png')
        self.task_name = task_name
        self.label = pyglet.text.Label(
            task_name,
            font_name='Times New Roman',
            font_size=24,
            x=30, y=self.window.height - 30,
            anchor_x='left', anchor_y='top'
        )
        self.window.on_draw = self.on_draw
        self.world = world
        self.viewpoint_car = None if len(self.world.cars) == 0 else self.world.cars[0]

    def draw_lane_surface(self, lane):
        gl.glColor3f(0.4, 0.4, 0.4)
        W = 1000
        graphics.draw(4, gl.GL_QUAD_STRIP, ('v2f',
                                            np.hstack([lane.p - lane.m * W - 0.5 * lane.w * lane.n,
                                                       lane.p - lane.m * W + 0.5 * lane.w * lane.n,
                                                       lane.q + lane.m * W - 0.5 * lane.w * lane.n,
                                                       lane.q + lane.m * W + 0.5 * lane.w * lane.n])
                                            ))

    def draw_lane_lines(self, lane):
        gl.glColor3f(1., 1., 1.)
        W = 1000
        graphics.draw(4, gl.GL_LINES, ('v2f',
                                       np.hstack([lane.p - lane.m * W - 0.5 * lane.w * lane.n,
                                                  lane.p + lane.m * W - 0.5 * lane.w * lane.n,
                                                  lane.p - lane.m * W + 0.5 * lane.w * lane.n,
                                                  lane.p + lane.m * W + 0.5 * lane.w * lane.n])
                                       ))

    def draw_car(self, x, color='yellow', opacity=255):
        sprite = self.sprites[color]
        sprite.x, sprite.y = x[0], x[1]
        sprite.rotation = -x[3] * 180. / np.pi
        sprite.opacity = opacity
        sprite.draw()

    def draw_object(self, obj):
        sprite = self.obj_sprites[obj.name]
        sprite.x, sprite.y = obj.x[0], obj.x[1]
        sprite.rotation = obj.x[2] if len(obj.x) >= 3 else 0.
        sprite.draw()

    def center(self):
        if self.viewpoint_car is not None:
            return np.array(self.viewpoint_car.trajectory[-1][0:2])
        else:
            return np.array([0., 0.])

    def camera(self):
        o = self.center()
        gl.glOrtho(o[0] - 1. / self.magnify, o[0] + 1. / self.magnify, o[1] - 1. / self.magnify,
                   o[1] + 1. / self.magnify, -1., 1.)

    def reset(self):
        for car in self.world.cars:
            car.reset()

    def on_draw(self):
        self.window.clear()
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glPushMatrix()
        gl.glLoadIdentity()
        self.camera()
        gl.glEnable(self.background.target)
        gl.glEnable(gl.GL_BLEND)
        gl.glBindTexture(self.background.target, self.background.id)
        W = 10000.
        graphics.draw(4, gl.GL_QUADS,
                      ('v2f', (-W, -W, W, -W, W, W, -W, W)),
                      ('t2f', (0., 0., W * 5., 0., W * 5., W * 5., 0., W * 5.))
                      )
        gl.glDisable(self.background.target)
        for lane in self.world.lanes:
            self.draw_lane_surface(lane)
            self.draw_lane_lines(lane)
        for obj in self.world.objects:
            self.draw_object(obj)
        for car in self.world.cars:
            self.draw_car(car.trajectory[-1], car.color)
        gl.glPopMatrix()

        self.label.text = self.task_name
        self.label.draw()
        self.iter +=1
        if self.iter%10 == 0:
            print('Iterations: ', self.iter)
        if self.iter == self.max_iters:
            self.event_loop.exit()

    def simulation_step(self, _=None):
        for car in self.world.cars:
            car.step()

    def run(self, reset=True):
        if reset:
            self.reset()
        pyglet.clock.schedule_interval(self.simulation_step, self.dt)
        self.event_loop.run()

    def exit_simulation(self):
        self.window.close()













import numpy as np

try:
    import gym
except ModuleNotFoundError:
    import sys
    sys.exit('This functionality requires OpenAI-gym to be installed')

try:
    from baselines.run import *
except ModuleNotFoundError:
    import sys
    sys.exit('This functionality requires baselines to be installed: Try installing from source')

try:
    import tensorflow as tf
except ModuleNotFoundError:
    import sys
    sys.exit('This functionality requires tensorflow to be installed')




class control_task:
    def __init__(self, baselines_params):
        self.env_id = baselines_params.env_id \
            if 'env_id' in baselines_params else 'CartPole-v1'
        self.alg = baselines_params.alg \
            if 'alg' in baselines_params else 'ppo2'
        self.num_timesteps = baselines_params.num_timesteps \
            if 'num_timesteps' in baselines_params else 1e6
        self.seed = baselines_params.seed \
                if 'seed' in baselines_params else None
        self.save_path = baselines_params.save_path \
                if 'save_path' is baselines_params else None
        self.network = baselines_params.network \
                if 'network' is baselines_params else None

        self.time_horizon = baselines_params.time_horizon \
            if 'time_horizon' in baselines_params else None

        self.env = gym.make(id=self.env_id)
        self.env.reset()
        self.model = None
        self.init_state = np.zeros(self.env.observation_space.shape)
        self.run_task = self.run_task_wo_retrain

    def build_args(self):
        self.args = ['', '--env=' + str(self.env_id), '--alg=' + str(self.alg), \
                     '--num_timesteps=' + str(self.num_timesteps)]
        if self.seed is not None:
            self.args.append('--seed=' + str(self.seed))
        if self.save_path is not None:
            self.args.append('--seed=' + str(self.save_path))
        if self.network is not None:
            self.args.append('--network='+str(self.network))

    def train(self):
        print("Training ", self.env_id, " with ", self.alg, " with ", self.num_timesteps)
        self.build_args()
        self.model = main(self.args)


    def play(self):
        state = self.model.initial_state if hasattr(self.model, 'initial_state') else None
        dones = np.zeros((1,))
        self.env.env.state = self.init_state
        obs = self.env.env.state
        self.trajectory = [obs]
        print("Initial state ", self.init_state)
        i = 0
        while True:
            i += 1
            obs_a = np.array(obs).reshape(1, len(obs))
            if state is not None:
                actions, _, state, _ = self.model.step(obs_a,S=state, M=dones)
            else:
                actions, _, _, _ = self.model.step(obs_a)
            obs, _, done, _ = self.env.step(actions[0])
            self.trajectory.append(obs)
            self.env.render()
            done = self.stop_simulation(done, i)
            if done:
                break
        self.env.close()

    def stop_simulation(self, done, i):
        if self.time_horizon is None:
            return done.any() if isinstance(done, np.ndarray) else done
        else:
            return i == self.time_horizon

    def use_sample(self, sample):
        raise NotImplementedError("Define in child task")

    def trajectory_definition(self):
        raise NotImplementedError("Define in child task")

    def run_task_retrain(self, sample):
        with tf.Session() as sess:
            self.env.reset()
            self.use_sample(sample)
            self.play()
            sess.close()
        tf.reset_default_graph()
        return self.trajectory_definition()

    def run_task_wo_retrain(self, sample):
        self.env.reset()
        self.use_sample(sample)
        self.play()
        return self.trajectory_definition()


from abc import ABC, abstractmethod
import socket
import dill


class Client(ABC):
    """Generic client for running simulations based on samples from the server.

    Users must implement the abstract method `simulate` to run a simulation.
    """

    def __init__(self, port, bufsize):
        self.port = port
        self.bufsize = bufsize

    def initialize(self):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.host = '127.0.0.1'
        try:
            self.socket.connect((self.host, self.port))
        except ConnectionResetError:
            self.close()
            return False
        except OSError as e:
            self.close()
            raise RuntimeError('unable to connect to server') from e
        return True

    def receive(self):
        data = []
        try:
            while True:
                msg = self.socket.recv(self.bufsize)
                if not msg:
                    break
                data.append(msg)
        except OSError:
            return False, None
        data = dill.loads(b"".join(data))
        return True, data

    def send(self, data):
        msg = dill.dumps(data)
        self.socket.send(msg)
        self.socket.shutdown(socket.SHUT_WR)

    def close(self):
        self.socket.close()

    def run_client(self):
        success = self.initialize()
        try:
            if success:
                success, sample = self.receive()
            if not success:
                print("No new sample received from server.")
                return False
            sim = self.simulate(sample)
            self.send(sim)
            return True
        finally:
            self.close()

    @abstractmethod
    def simulate(self, sample):
        """Run a simulation from the given sample.

        Returns:
            The outcome of the simulation (e.g. trajectories of objects), to be
            passed to the monitor.
        """
        pass



import carla
from agents.navigation.agent import *
from agents.navigation.controller import VehiclePIDController

import numpy as np
from collections import deque

from pid_agent import *

class PIDLocationController():
    def __init__(self, vehicle, clear_dist, K_P=1.0, K_D=0.0, K_I=0.0, dt=0.03):
        self._vehicle = vehicle
        self.clear_dist = clear_dist
        self._K_P = K_P
        self._K_D = K_D
        self._K_I = K_I
        self._dt = dt
        self._e_buffer = deque(maxlen=30)


    def run_step(self, location):
        '''Run step with target LOCATION.'''
        return self._pid_control(location, self._vehicle.get_transform())


    def _pid_control(self, target_location, cur_transform):
        _e = target_location.distance(cur_transform.location) - self.clear_dist
        self._e_buffer.append(_e)

        if len(self._e_buffer) >= 2:
            _de = (self._e_buffer[-1] - self._e_buffer[-2]) / self._dt
            _ie = sum(self._e_buffer) * self._dt
        else:
            _de = 0.0
            _ie = 0.0

        return (self._K_P * _e) + (self._K_D * _de / self._dt) + (self._K_I * _ie * self._dt)


class PIDFollowController(VehiclePIDController):
    '''Adds PID control for longitudinal position.'''
    def __init__(self, vehicle, clear_dist,
                 args_lateral={'K_P': 1.0, 'K_D': 0.0, 'K_I': 0.0},
                 args_longitudinal={'K_P': 1.0, 'K_D': 0.0, 'K_I': 0.0},
                 args_location={'K_P': 1.0, 'K_D': 0.0, 'K_I': 0.0}):
        super().__init__(vehicle, args_lateral=args_lateral,
                     args_longitudinal=args_longitudinal)
        self.loc_controller = PIDLocationController(
            vehicle, clear_dist, **args_location)


    def run_step(self, target_speed, waypoint, location=None):
        """
        Defaults to maintaining TARGET_SPEED toward WAYPOINT. If LOCATION,
        instead control throttle to minimize distance to LOCATION
        """
        control = super().run_step(target_speed, waypoint)
        if location:
            brake_throttle = self.loc_controller.run_step(location)
            if brake_throttle < 0:
                control.brake = np.clip(-brake_throttle, 0.0, 1.0)
                control.throttle = 0.0
            else:
                control.throttle = np.clip(brake_throttle, 0.0, 1.0)
                control.brake = 0.0

        return control
        



from verifai.client import Client

class ClientCar(Client):

    def __init__(self, simulation_data):
        port = simulation_data.port
        bufsize = simulation_data.bufsize
        super().__init__(port, bufsize)
        self.simulation = simulation_data.simulation

    def simulate(self, sample):
        return self.simulation(sample)






from verifai.simulators.car_simulator.examples.control_utils.LQR_computation import *
from verifai.simulators.car_simulator.simulator import *
from verifai.simulators.car_simulator.lane import *
from verifai.simulators.car_simulator.car_object import *
from verifai.simulators.car_simulator.client_car_sim import *
import numpy as np
from dotmap import DotMap

def controller(x_trajectory, u_trajectory, control_params):
    x, y, v, theta = x_trajectory[-1]
    wheelbase = control_params.wheelbase
    a_star = control_params.a_star
    v_star = control_params.v_star
    control_freq = control_params.control_freq
    dt = control_params.dt
    Q = control_params.Q
    R = control_params.R
    A, B = extract_AB(speed=v, dt=dt, wheelbase=wheelbase)
    if len(u_trajectory)%control_freq == 0:
        a = a_star if np.linalg.norm(v - v_star) > 0.1 else 0.0
        c = np.array([0.0, a * control_freq * dt, 0.0])
        K, k = discrete_LQR(A, B, Q, R, c)
        u = K.dot(np.array([[x],
                            [v - v_star],
                            [-theta+np.pi/2]])) + k
        u = min(float(u), np.pi / 4.)
        u = max(float(u), -np.pi / 4.)
        control = np.array([u, a])
    else:
        control = u_trajectory[-1]
    return control


def lanekeeping(sample, control_params, width=0.13):
    x_init = sample.init_conditions.x_init[0]
    theta_init = -sample.init_conditions.theta_init[0] + np.pi/2
    v_init = 0.0
    y_init = 0.0
    x0 = np.array([x_init, y_init, v_init, theta_init])
    a_star = control_params.a_star

    u_domain = {'omega':[-np.pi/4, np.pi/4],
                'acc':[-a_star, a_star]}
    compute_control = lambda x, u: controller(x_trajectory=x, u_trajectory=u,
                                              control_params=control_params)

    car = bicycle_model(x0=x0, u_domain=u_domain, compute_control=compute_control,
                        wheelbase = control_params.wheelbase, dt = control_params.dt,
                        color='red')

    lanes = []
    lanes.append(straight_lane([0., -1.], [0., 1.], width))
    lanes.append(lanes[0].shifted(1))
    lanes.append(lanes[0].shifted(-1))

    world = simulator_world(lanes=lanes, cars=[car])

    sim = simulator(dt=control_params.dt, iters = 100, sprite_scale=control_params.sprite_scale,
                    window_size = control_params.window_size, magnify =0.25)
    sim.init_simulator(world=world, task_name='Lane Keeping')
    sim.run()
    sim.exit_simulation()
    traj_x, traj_y, _, _ = np.array(car.trajectory).T
    data = {}
    traj = {}
    traj['xdeviation'] = [(j * control_params.dt, 0.5 - np.abs(v)) for j, v in enumerate(traj_x)]
    return traj

def lanekeeping_simulator(sample):
    print(sample)
    control_params = DotMap()
    control_params.wheelbase =2.995
    control_params.a_star = 3.968
    control_params.v_star = sample.init_conditions.cruising_speed[0]*5./18.
    control_params.dt = 0.032
    control_params.control_freq = 2
    control_params.R = 50*np.identity(1)
    control_params.Q= np.diag([100.0, 0.0, 5.0])
    control_params.sprite_scale = 1/800
    control_params.window_size = 800
    width = 1.0
    return lanekeeping(sample, control_params, width)


PORT = 8888
BUFSIZE = 4096
N_SIM_STEPS = 100
simulation_data = DotMap()
simulation_data.port = PORT
simulation_data.bufsize = BUFSIZE
simulation_data.simulation = lanekeeping_simulator

client_task = ClientCar(simulation_data)
while True:
    if not client_task.run_client():
        print("End of all simulations")
        break












from verifai.samplers.domain_sampler import BoxSampler
import numpy as np

class DistBayesOptSampler(BoxSampler):
    def __init__(self, domain, distBOparams):
        try:
            import GPyOpt
        except ModuleNotFoundError:
            raise RuntimeError(
                'DistBayesOptSampler requires GPyOpt to be installed')

        super().__init__(domain)
        self.num_GPs = distBOparams.num_preds
        self.func = distBOparams.func
        self.init_num = distBOparams.init_num
        if self.init_num < 1:
            raise RuntimeError(
                'init_num for DistBayesOptSampler must be at least 1')
        self.k = 10 if 'k' not in distBOparams else distBOparams.k
        self.bounds = []
        for i in range(self.dimension):
            self.bounds.append({'name': 'x_' + str(i), 'type': 'continuous',
                                'domain': (0, 1)})
        self.X = np.empty((0, self.dimension))
        self.Y = [np.empty((0, 1)) for i in range(self.num_GPs)]

    def getVector(self):
        import GPyOpt   # do this here to avoid slow import when unused
        import GPy

        if len(self.X) < self.init_num:
            # Do random sampling
            sample = np.random.uniform(0, 1, self.dimension)
        else:
            # Do BO
            X_vector = np.empty((self.num_GPs, 1))
            for i in range(self.num_GPs):
                BO = GPyOpt.methods.BayesianOptimization(f=None, batch_size=1,
                    domain=self.bounds, X=self.X, Y=self.Y[i], normalize_Y=False)
                X_vector[i] = BO.suggest_next_locations()[0]
            vals = []
            for Yi in self.Y:
                GP = GPy.models.GPRegression(self.X, Yi,
                    kernel=GPy.kern.Matern52(self.dimension))
                GP.optimize()
                m, v = GP.predict(X_vector)
                vals.append(m - self.k*v)
            func_val = self.func(vals)
            i = func_val.argmin()
            sample = X_vector[i]
        return tuple(sample), None

    def updateVector(self, vector, info, rho):
        assert len(rho) == self.num_GPs
        for i, sfb in enumerate(rho):
            self.Y[i] = np.vstack((self.Y[i], np.atleast_2d(sfb)))



import math
import numpy as np

EARTH_RADIUS = 6371008.0


def euclidean_dist(desired_x_y, comparative_x_y):
    d_x, d_y = desired_x_y[0], desired_x_y[1]
    c_x, c_y = comparative_x_y[0], comparative_x_y[1]
    return math.hypot(d_x - c_x, d_y - c_y)


def quaternion_for(theta, phi, psi):
    # (conversion math from http://www.xsquawkbox.net/xpsdk/mediawiki/MovingThePlane)
    sin_theta, cos_theta = math.sin(theta/2), math.cos(theta/2)
    sin_phi, cos_phi = math.sin(phi/2), math.cos(phi/2)
    sin_psi, cos_psi = math.sin(psi/2), math.cos(psi/2)
    quaternion = [
        cos_psi * cos_theta * cos_phi + sin_psi * sin_theta * sin_phi,
        cos_psi * cos_theta * sin_phi - sin_psi * sin_theta * cos_phi,
        cos_psi * sin_theta * cos_phi + sin_psi * cos_theta * sin_phi,
        -cos_psi * sin_theta * sin_phi + sin_psi * cos_theta * cos_phi
    ]
    return quaternion


def compute_heading_error(desired, real):
    '''
    param desired: desired heading
    param real: current heading
    return: heading error
    '''
    error, sign = desired - real, 1
    phi = abs(error) % 360
    if not (0 <= error <= 180 or -360 <= error <= -180):
        sign = -1
    if phi > 180:
        return (360 - phi) * sign
    return phi * sign


def cross_track_distance(start_lat, start_lon, end_lat, end_lon, d_lat, d_lon):
    # ref: http://www.movable-type.co.uk/scripts/latlong.html
    d13 = great_circle_distance_haversine(start_lat, start_lon, d_lat, d_lon) / EARTH_RADIUS
    theta12 = np.radians(initial_bearing(start_lat, start_lon, end_lat, end_lon))
    theta13 = np.radians(initial_bearing(start_lat, start_lon, d_lat, d_lon))
    return -(np.arcsin(np.sin(d13) * np.sin(theta13 - theta12)) * EARTH_RADIUS)


def great_circle_distance_haversine(start_lat, start_lon, end_lat, end_lon):
    phi1, phi2 = np.radians(start_lat), np.radians(end_lat)
    dphi, dlambda = np.radians(start_lat - end_lat), np.radians(start_lon - end_lon)
    a = np.sin(dphi / 2) * np.sin(dphi / 2) + np.cos(phi1) * np.cos(phi2) * \
                np.sin(dlambda / 2) * np.sin(dlambda / 2)
    c = 2 * np.arctan2(np.sqrt(a), np.sqrt(1 - a))
    return c * EARTH_RADIUS


def initial_bearing(start_lat, start_lon, end_lat, end_lon):
    lat1, lat2 = np.radians(start_lat), np.radians(end_lat)
    lon1, lon2 = np.radians(start_lon), np.radians(end_lon)
    y = np.sin(lon2 - lon1) * np.cos(lat2)
    x = np.cos(lat1) * np.sin(lat2) - np.sin(lat1) * \
                np.cos(lat2) * np.cos(lon2 - lon1)
    return np.degrees(np.arctan2(y, x)) % 360



"""Assorted utility functions"""

### Exceptions

class RejectionException(Exception):
    """Exception raised during sampling if the sample should be rejected."""
    pass

### Python conveniences

def cached(oldMethod):
    """Decorator for making a method with no arguments cache its result"""
    storageName = f'_cached_{oldMethod.__name__}'
    def newMethod(self):
        if not hasattr(self, storageName):
            setattr(self, storageName, oldMethod(self))
        return getattr(self, storageName)
    return newMethod

def cached_property(oldMethod):
    """Decorator for making a property which caches its value"""
    return property(cached(oldMethod))



from abc import ABC
import networkx as nx
import mtl

class specification_monitor(ABC):
    def __init__(self, specification):
        self.specification = specification

    def evaluate(self, traj):
        return self.specification(traj)


class mtl_specification(specification_monitor):
    def __init__(self, specification):
        mtl_specs = [mtl.parse(spec) for spec in specification]
        mtl_spec = mtl_specs[0]
        if len(mtl_specs) > 1:
            for spec in mtl_specs[1:]:
                mtl_spec = (mtl_spec & spec)
        super().__init__(mtl_spec)

    def evaluate(self, traj):
        return self.specification(traj)

class multi_objective_monitor(specification_monitor):
    def __init__(self, specification, priority_graph=None, linearize=False):
        super().__init__(specification)
        self.linearize = linearize
        if priority_graph is None:
            self.graph = nx.DiGraph()
            self.graph.add_nodes_from(range(self.num_objectives))
        else:
            self.graph = priority_graph
        if linearize: # "linearize" the prioritize graph by topologically sorting and connecting nodes
            self._linearize()
    
    def _linearize(self):
        new_graph = nx.DiGraph()
        S = set([node for node, degree in self.graph.in_degree() if degree == 0])
        nodes = []
        while len(S) > 0:
            n = random.choice(S)
            S.remove(n)
            nodes.append(n)
            neighbors = self.graph.neighbors(n)
            random.shuffle(neighbors)
            for m in neighbors:
                self.graph.remove_edge(n, m)
                if self.graph.in_degree(m) == 0:
                    S.add(m)
        new_graph.add_nodes_from(nodes)
        for i in range(1, len(nodes)):
            new_graph.add_edge(nodes[i - 1], nodes[i])
        self.graph = new_graph


from verifai.client import Client

class ClientWebots(Client):

    def __init__(self, simulation_data):
        port = simulation_data.port
        bufsize = simulation_data.bufsize
        super().__init__(port, bufsize)
        self.task = simulation_data.task

    def simulate(self, sample):
        return self.task.run_task(sample)

    def run_client(self):
        success = super().run_client()
        if success:
            self.task.close()



"""Random sampling"""

from verifai.samplers.domain_sampler import DomainSampler

class RandomSampler(DomainSampler):
    """Samples uniformly at random or from a given distribution"""

    def __init__(self, domain, distribution=None):
        super().__init__(domain)
        self.distribution = distribution

    def getSample(self):
        if self.distribution is None:
            return self.domain.uniformPoint(), None
        else:
            return self.distribution.sample(), None

    def __repr__(self):
        rep = f'RandomSampler({self.domain}'
        if self.distribution is not None:
            rep += f', distribution={self.distribution}'
        return rep + ')'



from abc import ABC
from verifai.server import Server, ParallelServer
from verifai.scenic_server import ScenicServer, ParallelScenicServer
from verifai.samplers import TerminationException
from dotmap import DotMap
from verifai.monitor import mtl_specification, specification_monitor, multi_objective_monitor
from verifai.error_table import error_table
import numpy as np
import progressbar
from statsmodels.stats.proportion import proportion_confint
import time

def parallelized(server_class):
    if server_class == Server:
        return ParallelServer
    elif server_class == ScenicServer:
        return ParallelScenicServer

class falsifier(ABC):
    def __init__(self, monitor, sampler_type=None, sampler=None, sample_space=None,
                 falsifier_params=None, server_options={}, server_class=Server):
        self.sample_space = sample_space
        self.sampler_type = sampler_type
        self.sampler = sampler
        self.samples = {}
        self.monitor = monitor

        params = DotMap(
            save_error_table=True, save_safe_table=True,
            error_table_path=None, safe_table_path=None,
            n_iters=1000, ce_num_max=np.inf, fal_thres=0,
            max_time=None,
            sampler_params=None, verbosity=0,
        )
        if falsifier_params is not None:
            params.update(falsifier_params)
        if params.sampler_params is None:
            params.sampler_params = DotMap(thres=params.fal_thres)
        self.multi = isinstance(self.monitor, multi_objective_monitor)
        if self.multi:
            params.sampler_params.priority_graph = self.monitor.graph
        self.save_error_table = params.save_error_table
        self.save_safe_table = params.save_safe_table
        self.error_table_path = params.error_table_path
        self.safe_table_path = params.safe_table_path
        if not hasattr(self, 'num_workers'):
            self.num_workers = 1
        self.n_iters, self.ce_num_max = params.n_iters, params.ce_num_max
        self.max_time = params.max_time
        self.fal_thres = params.fal_thres
        self.sampler_params = params.sampler_params
        self.verbosity = params.verbosity

        server_params = DotMap(init=True)
        if server_options is not None:
            server_params.update(server_options)
        if server_params.init:
            self.init_server(server_params, server_class)
            self.init_error_table()

    def init_server(self, server_options, server_class):
        if self.verbosity >= 1:
            print("Initializing server")
        sampling_data = DotMap()
        if self.sampler_type is None:
            self.sampler_type = 'random'
        sampling_data.sampler_type = self.sampler_type
        sampling_data.sample_space = self.sample_space
        sampling_data.sampler_params = self.sampler_params
        sampling_data.sampler = self.sampler

        self.server = server_class(sampling_data, self.monitor, options=server_options)
        if self.verbosity >= 1:
            print("Server ready")

    def init_error_table(self):
        # Initializing error table
        if self.save_error_table:
            self.error_table = error_table(space = self.server.sample_space)
        if self.save_safe_table:
            self.safe_table = error_table(space = self.server.sample_space)

    def populate_error_table(self, sample, rho, error=True):
        if error:
            self.error_table.update_error_table(sample, rho)
            if self.error_table_path:
                self.write_table(self.error_table.table, self.error_table_path)
        else:
            self.safe_table.update_error_table(sample, rho)
            if self.safe_table_path:
                self.write_table(self.safe_table.table, self.safe_table_path)

    @staticmethod
    def write_table(table, path):
        if len(table) <= 1:
            table.to_csv(path)
        else:
            table.tail(1).to_csv(path, mode='a', header=False)

    def analyze_error_table(self, analysis_params= None, error=None):
        if self.save_error_table and (error is None or error is True):
            self.error_analysis = self.error_table.analyze(analysis_params)

            if 'k_closest' in self.error_analysis:
                self.error_analysis.k_closest_samples = \
                    [self.samples[i] for i in self.error_analysis.k_closest]
            if 'random' in self.error_analysis:
                self.error_analysis.random_samples = \
                    [self.samples[i] for i in self.error_analysis.random]
        if self.save_safe_table and (error is None or error is False):
            self.safe_analysis = self.safe_table.analyze(analysis_params)

            if 'k_closest' in self.safe_analysis:
                self.safe_analysis.k_closest_samples = \
                    [self.samples[i] for i in self.safe_analysis.k_closest]
            if 'random' in self.safe_analysis:
                self.safe_analysis.random_samples = \
                    [self.samples[i] for i in self.safe_analysis.random]

    def get_confidence_interval(self, confidence_level=0.95):
        N = len(self.samples)
        c = len(self.error_table.table)
        return proportion_confint(c, N, alpha=1 - confidence_level, method='beta')

    def run_falsifier(self):
        i = 0
        ce_num = 0
        server_samples = []
        rhos = []
        self.total_sample_time = 0
        self.total_simulate_time = 0
        if self.verbosity >= 1:
            suffix = ''
            if self.n_iters:
                suffix = f' for {self.n_iters} iterations'
            if self.max_time:
                suffix += f' for {self.max_time:.0f} seconds'
            print('Running falsification' + suffix)
        if self.verbosity >= 2:
            print(f'Server class is {type(self.server)}')

        if self.verbosity >= 1:
            if self.n_iters is not None:
                bar = progressbar.ProgressBar(max_value=self.n_iters)
            else:
                widgets = ['Samples generated: ', progressbar.Counter('%(value)d'),
                ' (', progressbar.Timer(), ')']
                bar = progressbar.ProgressBar(widgets=widgets)

        try:
            while True:
                try:
                    sample, rho, timings = self.server.run_server()
                    self.total_sample_time += timings.sample_time
                    self.total_simulate_time += timings.simulate_time
                except TerminationException:
                    if self.verbosity >= 1:
                        print("Sampler has generated all possible samples")
                    break
                if self.verbosity >= 2:
                    print("Sample no: ", i, "\nSample: ", sample, "\nRho: ", rho)
                self.samples[i] = sample
                server_samples.append(sample)
                rhos.append(rho)
                i += 1
                if self.verbosity >= 1:
                    bar.update(i)
                if i == 1:
                    t0 = time.time()
                if self.n_iters is not None and i == self.n_iters:
                    break
                if self.max_time is not None and time.time() - t0 >= self.max_time:
                    break
        finally:
            if self.verbosity >= 1:
                bar.finish()
            self.server.terminate()
        for sample, rho in zip(server_samples, rhos):
            ce = any([r <= self.fal_thres for r in rho]) if self.multi else rho <= self.fal_thres
            if ce:
                if self.save_error_table:
                    self.populate_error_table(sample, rho)
                ce_num = ce_num + 1
                if ce_num >= self.ce_num_max:
                    break
            elif self.save_safe_table:
                self.populate_error_table(sample, rho, error=False)
        if self.verbosity >= 1:
            print('Falsification complete.')

class generic_falsifier(falsifier):
    def __init__(self,  monitor=None, sampler_type= None, sample_space=None, sampler=None,
                 falsifier_params=None, server_options={}, server_class=Server):
        if monitor is None:
            class monitor(specification_monitor):
                def __init__(self):
                    def specification(traj):
                        return np.inf
                    super().__init__(specification)
            monitor = monitor()

        super().__init__(sample_space=sample_space, sampler_type=sampler_type,
                         monitor=monitor, falsifier_params=falsifier_params, sampler=sampler,
                         server_options=server_options, server_class=server_class)

class mtl_falsifier(generic_falsifier):
    def __init__(self, specification, sampler_type = None, sample_space=None, sampler=None,
                 falsifier_params=None, server_options={}, server_class=Server):
        monitor = mtl_specification(specification=specification)
        super().__init__(sample_space=sample_space, sampler_type=sampler_type,
                         monitor=monitor, falsifier_params=falsifier_params, sampler=sampler,
                         server_options=server_options, server_class=server_class)

class parallel_falsifier(falsifier):

    def __init__(self, monitor, sampler_type=None, sample_space=None,
                 falsifier_params=None, server_options={}, server_class=Server, 
                 sampler=None):
        self.num_workers = server_options.num_workers
        self.scenic_path = server_options.scenic_path
        super().__init__(sample_space=sample_space, sampler_type=sampler_type,
                         monitor=monitor, falsifier_params=falsifier_params, sampler=sampler,
                         server_options=server_options, server_class=parallelized(server_class))

class generic_parallel_falsifier(parallel_falsifier):
    def __init__(self, monitor=None, sampler_type= None, sample_space=None, sampler=None,
                 falsifier_params=None, server_options={}, server_class=Server):
        if monitor is None:
            class monitor(specification_monitor):
                def __init__(self):
                    def specification(traj):
                        return np.inf
                    super().__init__(specification)
            monitor = monitor()
        self.scenario_params = server_options.scenario_params

        super().__init__(sample_space=sample_space, sampler_type=sampler_type,
                         monitor=monitor, falsifier_params=falsifier_params,
                         server_options=server_options, server_class=server_class,
                         sampler=sampler)

    def init_server(self, server_options, server_class):
        if self.verbosity >= 1:
            print("Initializing server")
        sampling_data = DotMap()
        if self.sampler_type is None:
            self.sampler_type = 'random'
        sampling_data.sampler_type = self.sampler_type
        sampling_data.sample_space = self.sample_space
        sampling_data.sampler_params = self.sampler_params

        self.server = server_class(self.num_workers, self.n_iters, sampling_data, self.scenic_path,
        self.monitor, options=server_options, max_time=self.max_time, sampler=self.sampler)

    def run_falsifier(self):
        i = 0
        ce_num = 0
        try:
            outputs = self.server.run_server()
        finally:
            self.server.terminate()
        samples, rhos = zip(*outputs)
        if isinstance(self.monitor, multi_objective_monitor):
            counterexamples = [any([r <= self.fal_thres for r in rho]) for rho in rhos]
        else:
            counterexamples = [r <= self.fal_thres for r in rhos]
        for i, (sample, ce, rho) in enumerate(zip(samples, counterexamples, rhos)):
            if self.verbosity >= 1:
                print("Sample no: ", i, "\nSample: ", sample, "\nRho: ", rho)
            self.samples[i] = sample
            if ce:
                if self.save_error_table:
                    self.populate_error_table(sample, rho)
                ce_num = ce_num + 1
                if ce_num >= self.ce_num_max:
                    break
            elif self.save_safe_table:
                self.populate_error_table(sample, rho, error=False)




"""Interface to Scenic."""

import importlib.metadata

import scenic
from scenic.core.distributions import needsSampling, Options
from scenic.core.vectors import Vector
from scenic.core.type_support import canCoerceType, coerce, underlyingType

# TODO unify handling of these custom types!
from scenic.simulators.utils.colors import Color
from scenic.simulators.gta.interface import CarModel as GTACarModel
from scenic.simulators.webots.road.car_models import (
    CarModel as WebotsCarModel, carModels as webotsCarModels)

from verifai.features import (Constant, Categorical, Real, Box, Array, Struct,
                              Feature, FeatureSpace)
from verifai.samplers.feature_sampler import FeatureSampler
from verifai.utils.frozendict import frozendict

scenicMajorVersion = int(importlib.metadata.version('scenic').split('.')[0])
if scenicMajorVersion >= 3:
    from scenic.core.vectors import Orientation
else:
    Orientation = None

scalarDomain = Real()
vectorDim = 2 if scenicMajorVersion < 3 else 3
vectorDomain = Array(scalarDomain, (vectorDim,))
orientationDomain = Array(scalarDomain, (4,))  # as quaternions
gtaModelDomain = Categorical(*GTACarModel.models.values())
webotsModelDomain = Categorical(*webotsCarModels)
colorDomain = Box((0, 1), (0, 1), (0, 1))

def convertToVerifaiType(value, strict=True):
    """Attempt to convert a Scenic value to a type known to VerifAI"""
    ty = underlyingType(value)
    if ty is float or ty is int:
        return float(value)
    elif ty is list or ty is tuple:
        return tuple(convertToVerifaiType(e, strict=strict) for e in value)
    elif issubclass(ty, dict) and not needsSampling(value):
        return frozendict(value)
    elif ty is GTACarModel:
        return value
    elif ty is WebotsCarModel:
        return value
    elif ty is Color:
        return value
    elif canCoerceType(ty, Vector):
        return tuple(coerce(value, Vector))
    elif Orientation and isinstance(value, Orientation):
        return tuple(value.getRotation().as_quat())
    elif strict:    # Unknown type, so give up if we're being strict
        raise RuntimeError(
            f'attempted to convert Scenic value {value} of unknown type {ty}')
    else:
        return value

def domainForValue(value):
    """Return a Domain for this type of Scenic value, when possible"""
    ty = underlyingType(value)
    if ty is float or ty is int:
        domain = scalarDomain
    elif ty is GTACarModel:
        domain = gtaModelDomain
    elif ty is WebotsCarModel:
        domain = webotsModelDomain
    elif ty is Color:
        domain = colorDomain
    elif canCoerceType(ty, Vector):
        domain = vectorDomain
    elif ty is Orientation:
        domain = orientationDomain
    elif ty is str:
        # We can only handle strings when they come from a finite set of
        # possibilities; we heuristically detect that here.
        if isinstance(value, Options):
            domain = Categorical(*value.options)
        else:
            domain = None   # we can't ensure the domain is finite
    else:
        domain = None   # no corresponding Domain known
    if not needsSampling(value):
        # We can handle constants of unknown types, but when possible we
        # convert the value to a VerifAI type.
        value = convertToVerifaiType(value, strict=False)
        return Constant(value)
    return domain

def pointForValue(dom, scenicValue):
    """Convert a sampled Scenic value to a point in the corresponding Domain"""
    if isinstance(dom, Constant):
        value = convertToVerifaiType(scenicValue, strict=False)
        assert value == dom.value
        return value
    elif isinstance(dom, Categorical):
        value = convertToVerifaiType(scenicValue, strict=False)
        assert value in dom.values
        return value
    elif dom == scalarDomain:
        if not isinstance(scenicValue, (float, int)):
            raise RuntimeError(
                f'could not coerce Scenic value {scenicValue} to scalar')
        return coerce(scenicValue, float)
    elif dom == vectorDomain:
        return tuple(coerce(scenicValue, Vector))
    elif dom == orientationDomain:
        return tuple(scenicValue.getRotation().as_quat())
    elif dom == colorDomain:
        assert isinstance(scenicValue, (tuple, list))
        assert len(scenicValue) == 3
        return scenicValue
    else:
        raise RuntimeError(
            f'Scenic value {scenicValue} has unexpected domain {dom}')

#: Scenic global parameters not included in generated samples
ignoredParameters = {
    'externalSampler', 'externalSamplerRejectionFeedback',
    'verifaiSamplerType', 'verifaiSamplerParams',
    'behavior',
}
#: Scenic object properties not included in generated samples
defaultIgnoredProperties = {
    'shape',
    'viewAngle', 'viewAngles', 'visibleDistance', 'cameraOffset',
    'viewRayDensity', 'viewRayCount', 'viewRayDistanceScaling',
    'baseOffset', 'contactTolerance', 'onDirection', 'sideComponentThresholds',
    'allowCollisions', 'requireVisible', 'regionContainedIn', 'occluding',
    'showVisibleRegion',
    'mutator', 'mutationEnabled', 'mutationScale',
    'positionStdDev', 'headingStdDev', 'orientationStdDev',
    'behavior', 'lastActions',
}
# certain built-in properties requiring type normalization before Scenic 3
normalizedProperties = {
    'position': Vector,
    'heading': float
} if scenicMajorVersion < 3 else {}
# hard-coded Domains for certain properties
specialDomainProperties = {
    'webotsType': Categorical(*(model.name for model in webotsCarModels)),
    'color': colorDomain,   # to allow Colors, tuples, or lists
}

def domainForObject(obj, ignoredProperties):
    """Construct a Domain for the given Scenic Object"""
    domains = {}
    for prop in obj.properties:
        if prop in ignoredProperties or prop.startswith('_'):
            continue
        value = getattr(obj, prop)
        if prop in normalizedProperties:
            value = coerce(value, normalizedProperties[prop])
        # TODO improve this system... (need to get better type info in Scenic)
        if prop in specialDomainProperties and needsSampling(value):
            dom = specialDomainProperties[prop]
        else:
            dom = domainForValue(value)
        if dom is None:
            ty = underlyingType(value)
            print(f'WARNING: skipping property "{prop}" of unknown type {ty}')
        else:
            domains[prop] = dom
    # add type as additional property
    value = type(obj).__name__
    dom = domainForValue(value)
    assert dom is not None
    assert 'type' not in domains
    domains['type'] = dom
    return Struct(domains)

def pointForObject(dom, obj):
    """Convert a sampled Scenic Object to a point in its Domain"""
    values = []
    for prop, subdom in dom.domainNamed.items():
        if prop == 'type':      # special case for 'type' pseudo-property
            scenicValue = type(obj).__name__
        elif (prop == 'speed'
              and isinstance(subdom, Constant)
              and scenicMajorVersion < 3):
            # work around bug in Scenic 2.1
            scenicValue = subdom.value
        else:
            scenicValue = getattr(obj, prop)
        values.append(pointForValue(subdom, scenicValue))
    return dom.makePoint(*values)

def spaceForScenario(scenario, ignoredProperties):
    """Construct a FeatureSpace for the given Scenic Scenario."""
    # create domains for objects
    assert scenario.egoObject is scenario.objects[0]
    doms = (domainForObject(obj, ignoredProperties)
            for obj in scenario.objects)
    objects = Struct({
        ScenicSampler.nameForObject(i): dom
        for i, dom in enumerate(doms)
    })

    # create domains for global parameters
    paramDoms = {}
    quotedParams = {}
    for param, value in scenario.params.items():
        if param in ignoredParameters:
            continue
        dom = domainForValue(value)
        if dom is None:
            ty = underlyingType(value)
            print(f'WARNING: skipping param "{param}" of unknown type {ty}')
        else:
            if not param.isidentifier():    # munge quoted parameter names
                newparam = 'quoted_param' + str(len(quotedParams))
                quotedParams[newparam] = param
                param = newparam
            paramDoms[param] = dom
    params = Struct(paramDoms)

    space = FeatureSpace({
        'objects': Feature(objects),
        'params': Feature(params)
    })
    return space, quotedParams

class ScenicSampler(FeatureSampler):
    """Samples from the induced distribution of a Scenic scenario.

    Created using the `fromScenario` and `fromScenicCode` class methods.

    See :ref:`scene generation` in the Scenic documentation for details of how
    Scenic's sampler works. Note that VerifAI's other samplers can be used from
    within a Scenic scenario by defining :term:`external parameters`.
    """

    def __init__(self, scenario, maxIterations=None, ignoredProperties=None):
        self.scenario = scenario
        self.maxIterations = 2000 if maxIterations is None else maxIterations
        self.lastScene = None
        if ignoredProperties is None:
            ignoredProperties = defaultIgnoredProperties
        space, self.quotedParams = spaceForScenario(scenario, ignoredProperties)
        super().__init__(space)

    @classmethod
    def fromScenario(cls, path, maxIterations=None,
                     ignoredProperties=None, **kwargs):
        """Create a sampler corresponding to a Scenic program.

        The only required argument is ``path``, and ``maxIterations`` may be useful if
        your scenario requires a very large number of rejection sampling iterations.
        See `scenic.scenarioFromFile` for details on optional keyword arguments used to
        customize compilation of the Scenic file.

        Args:
            path (str): path to a Scenic file.
            maxIterations (int): maximum number of rejection sampling iterations
              (default 2000).
            ignoredProperties: properties of Scenic objects to not include in
              generated samples (see ``defaultIgnoredProperties`` for the default).
            kwargs: additional keyword arguments passed to `scenic.scenarioFromFile`;
              e.g. ``params`` to override global parameters or ``model`` to set the
              :term:`world model`.
        """
        scenario = scenic.scenarioFromFile(path, **kwargs)
        return cls(scenario, maxIterations=maxIterations,
                   ignoredProperties=ignoredProperties)

    @classmethod
    def fromScenicCode(cls, code, maxIterations=None,
                       ignoredProperties=None, **kwargs):
        """As above, but given a Scenic program as a string."""
        scenario = scenic.scenarioFromString(code, **kwargs)
        return cls(scenario, maxIterations=maxIterations,
                   ignoredProperties=ignoredProperties)

    def nextSample(self, feedback=None):
        ret = self.scenario.generate(
            maxIterations=self.maxIterations, feedback=feedback, verbosity=0
        )
        self.lastScene, _ = ret
        return self.pointForScene(self.lastScene)

    def pointForScene(self, scene):
        """Convert a sampled Scenic :obj:`~scenic.core.scenarios.Scene` to a point in our feature space.

        The `FeatureSpace` used by this sampler consists of 2 features:

            * ``objects``, which is a `Struct` consisting of attributes ``object0``,
              ``object1``, etc. with the properties of the corresponding objects
              in the Scenic program. The names of these attributes may change in a
              future version of VerifAI: use the `nameForObject` function to
              generate them.
            * ``params``, which is a `Struct` storing the values of the
              :term:`global parameters` of the Scenic program (use
              `paramDictForSample` to extract them).
        """
        lengths, dom = self.space.domains
        assert lengths is None
        assert scene.egoObject is scene.objects[0]
        objDomain = dom.domainNamed['objects']
        assert len(objDomain.domains) == len(scene.objects)
        objects = {
            self.nameForObject(i):
            pointForObject(objDomain.domainNamed[self.nameForObject(i)], obj)
            for i, obj in enumerate(scene.objects)
        }
        objPoint = objDomain.makePoint(**objects)

        paramDomain = dom.domainNamed['params']
        params = {}
        for param, subdom in paramDomain.domainNamed.items():
            originalName = self.quotedParams.get(param, param)
            params[param] = pointForValue(subdom, scene.params[originalName])
        paramPoint = paramDomain.makePoint(**params)

        return self.space.makePoint(objects=objPoint, params=paramPoint)

    @staticmethod
    def nameForObject(i):
        """Name used in the `FeatureSpace` for the Scenic object with index i.

        That is, if ``scene`` is a :obj:`~scenic.core.scenarios.Scene`, the object
        ``scene.objects[i]``.
        """
        return f'object{i}'

    def paramDictForSample(self, sample):
        """Recover the dict of :term:`global parameters` from a `ScenicSampler` sample."""
        params = sample.params._asdict()
        corrected = {}
        for newName, quotedParam in self.quotedParams.items():
            corrected[quotedParam] = params.pop(newName)
        corrected.update(params)
        return corrected



import numpy as np
import networkx as nx
from itertools import product
from verifai.samplers.domain_sampler import BoxSampler, DiscreteBoxSampler, \
    DomainSampler, SplitSampler
from verifai.samplers.random_sampler import RandomSampler
from verifai.samplers.cross_entropy import DiscreteCrossEntropySampler
from verifai.samplers.multi_objective import MultiObjectiveSampler

class MultiArmedBanditSampler(DomainSampler):
    def __init__(self, domain, mab_params):
        super().__init__(domain)
        self.alpha = mab_params.alpha
        self.thres = mab_params.thres
        self.cont_buckets = mab_params.cont.buckets
        self.cont_dist = mab_params.cont.dist
        self.disc_dist = mab_params.disc.dist
        self.cont_ce = lambda domain: ContinuousMultiArmedBanditSampler(domain=domain,
                                                     buckets=self.cont_buckets,
                                                     dist=self.cont_dist,
                                                     alpha=self.alpha,
                                                     thres=self.thres)
        self.disc_ce = lambda domain: DiscreteMultiArmedBanditSampler(domain=domain,
                                                   dist=self.disc_dist,
                                                   alpha=self.alpha,
                                                   thres=self.thres)
        partition = (
            (lambda d: d.standardizedDimension > 0, self.cont_ce),
            (lambda d: d.standardizedIntervals, self.disc_ce)
        )
        self.split_sampler = SplitSampler.fromPartition(domain,
                                                        partition,
                                                        RandomSampler)
        self.cont_sampler, self.disc_sampler = None, None
        self.rand_sampler = None
        for subsampler in self.split_sampler.samplers:
            if isinstance(subsampler, ContinuousMultiArmedBanditSampler):
                assert self.cont_sampler is None
                if 'priority_graph' in mab_params:
                    subsampler.set_graph(mab_params.priority_graph)
                self.cont_sampler = subsampler
            elif isinstance(subsampler, DiscreteMultiArmedBanditSampler):
                assert self.disc_sampler is None
                self.disc_sampler = subsampler
            else:
                assert isinstance(subsampler, RandomSampler)
                assert self.rand_sampler is None
                self.rand_sampler = subsampler

    def getSample(self):
        return self.split_sampler.getSample()

    def update(self, sample, info, rho):
        self.split_sampler.update(sample, info, rho)

class ContinuousMultiArmedBanditSampler(BoxSampler, MultiObjectiveSampler):
    def __init__(self, domain, alpha, thres,
                 buckets=10, dist=None, restart_every=100):
        super().__init__(domain)
        if isinstance(buckets, int):
            buckets = np.ones(self.dimension) * buckets
        elif len(buckets) > 1:
            assert len(buckets) == self.dimension
        else:
            buckets = np.ones(self.dimension) * buckets[0]
        if dist is not None:
            assert (len(dist) == len(buckets))
        if dist is None:
            dist = np.array([np.ones(int(b))/b for b in buckets])
        self.buckets = buckets
        self.dist = dist
        self.alpha = alpha
        self.thres = thres
        self.current_sample = None
        self.counts = np.array([np.ones(int(b)) for b in buckets])
        self.errors = np.array([np.zeros(int(b)) for b in buckets])
        self.t = 1
        self.counterexamples = dict()
        self.is_multi = False
        self.invalid = np.array([np.zeros(int(b)) for b in buckets])
        self.monitor = None
        self.rho_values = []
        self.restart_every = restart_every

    def getVector(self):
        return self.generateSample()
    
    def generateSample(self):
        proportions = self.errors / self.counts
        Q = proportions + np.sqrt(2 / self.counts * np.log(self.t))
        # choose the bucket with the highest "goodness" value, breaking ties randomly.
        bucket_samples = np.array([np.random.choice(np.flatnonzero(np.isclose(Q[i], Q[i].max())))
            for i in range(len(self.buckets))])
        self.current_sample = bucket_samples
        ret = tuple(np.random.uniform(bs, bs+1.)/b for b, bs
              in zip(self.buckets, bucket_samples))
        return ret, bucket_samples
    
    def updateVector(self, vector, info, rho):
        assert rho is not None
        self.t += 1
        # "random restarts" to generate a new topological sort of the priority graph
        # every restart_every samples.
        if self.is_multi:
            if self.monitor is not None and self.monitor.linearize and t % self.restart_every == 0:
                self.monitor._linearize()
            self.update_dist_from_multi(vector, info, rho)
            return
        for i, b in enumerate(info):
            self.counts[i][b] += 1.
            if rho < self.thres:
                self.errors[i][b] += 1.

    # is rho1 better than rho2?
    # partial pre-ordering on objective functions, so it is possible that:
    # is_better_counterexample(rho1, rho2)
    # and is_better_counterxample(rho2, rho1) both return False
    def is_better_counterexample(self, ce1, ce2):
        if ce2 is None:
            return True
        all_same = True
        already_better = [False] * self.num_properties
        for node in nx.dfs_preorder_nodes(self.priority_graph):
            if already_better[node]:
                continue
            b1 = ce1[node]
            b2 = ce2[node]
            all_same = all_same and b1 == b2
            if b2 and not b1:
                return False
            if b1 and not b2:
                already_better[node] = True
                for subnode in nx.descendants(self.priority_graph, node):
                    already_better[subnode] = True
        return not all_same

    def _get_total_counterexamples(self):
        return sum(self.counterexamples.values())
    
    @property
    def counterexample_values(self):
        return [ce in self.counterexamples for ce in self.rho_values]

    def _add_to_running(self, ce):
        if ce in self.counterexamples:
            return True
        to_remove = set()
        # if there is already a better counterexample, don't add this.
        if len(self.counterexamples) > 0:
            for other_ce in self.counterexamples:
                if self.is_better_counterexample(other_ce, ce):
                    return False
        # remove all worse counterexamples than this.
        for other_ce in self.counterexamples:
            if self.is_better_counterexample(ce, other_ce):
                to_remove.add(other_ce)
        for other_ce in to_remove:
            del self.counterexamples[other_ce]
        self.counterexamples[ce] = np.array([np.zeros(int(b)) for b in self.buckets])
        return True
    
    def update_dist_from_multi(self, sample, info, rho):
        try:
            iter(rho)
        except:
            for i, b in enumerate(info):
                self.invalid[i][b] += 1.
            return
        if len(rho) != self.num_properties:
            for i, b in enumerate(info):
                self.invalid[i][b] += 1.
            return
        # print('inside update_dist_from_multi')
        counter_ex = tuple(
            rho[node] < self.thres[node] for node in nx.dfs_preorder_nodes(self.priority_graph)
        )
        self.rho_values.append(counter_ex)
        # print(f'counter_ex = {counter_ex}')
        # print(self.counterexamples)
        is_ce = self._add_to_running(counter_ex)
        for i, b in enumerate(info):
            self.counts[i][b] += 1.
            if is_ce:
                self.counterexamples[counter_ex][i][b] += 1.
        self.errors = self.invalid + self._get_total_counterexamples()

class DiscreteMultiArmedBanditSampler(DiscreteCrossEntropySampler):
    pass


# This file defines the car objects
from verifai.simulators.car_simulator.car_dynamics import car_dynamics
import numpy as np

class car:
    def __init__(self, x0, u_domain, color='yellow'):
        self.x0 = x0
        self.u_domain = u_domain
        self.color = color
        self.trajectory = [x0]
    def reset(self):
        self.trajectory = [self.x0]
        self.control_trajectory = []

class bicycle_model(car):
    def __init__(self, x0, u_domain, compute_control,
                 wheelbase=3.0, dt=0.1, color='yellow'):
        if u_domain is None:
            u_domain= {'omega':[-np.pi/4, np.pi/4], 'acc':[-4.0, 4.0]}
        elif 'omega' not in u_domain:
            u_domain['omega'] = [-np.pi/4, np.pi/4]
        elif 'acc' not in u_domain:
            u_domain['acc'] = [-4.0, 4.0]
        super().__init__(x0, u_domain, color)
        self.dynamics = car_dynamics(wheelbase=wheelbase, dt=dt)
        self.control_trajectory = []
        self.compute_control = compute_control

    def assert_control(self, u):
        u[0] = min(u[0], self.u_domain['omega'][1])
        u[0] = max(u[0], self.u_domain['omega'][0])

        u[1] = min(u[1], self.u_domain['acc'][1])
        u[1] = max(u[1], self.u_domain['acc'][0])

        return u

    def step(self):
        x = self.trajectory[-1]
        u = self.compute_control(self.trajectory, self.control_trajectory)
        u = self.assert_control(u)
        self.control_trajectory.append(u)
        self.trajectory.append(self.dynamics.next_state(x, u))




from dataclasses import dataclass
import socket
import time

import dill
from dotmap import DotMap

from verifai.features.features import *
from verifai.samplers.feature_sampler import *

def choose_sampler(sample_space, sampler_type,
                   sampler_params=None):
    if sampler_type == 'random':
        return 'random', FeatureSampler.randomSamplerFor(sample_space)

    if sampler_type == 'grid':
        return 'grid', FeatureSampler.gridSamplerFor(sample_space)

    if sampler_type == 'halton':
        if sampler_params is None:
            halton_params = default_sampler_params('halton')
        else:
            halton_params = default_sampler_params('halton')
            halton_params.update(sampler_params)

        sampler = FeatureSampler.haltonSamplerFor(sample_space,
                                                  halton_params=halton_params)
        return 'halton', sampler
    if sampler_type == 'ce':
        if sampler_params is None:
            ce_params = default_sampler_params('ce')
        else:
            ce_params = default_sampler_params('ce')
            if 'cont' in sampler_params:
                if 'buckets' in sampler_params.cont:
                    ce_params.cont.buckets = sampler_params.cont.buckets
                if 'dist' in sampler_params.cont:
                    ce_params.cont.dist = sampler_params.cont.dist
            if 'dist' in sampler_params.disc:
                ce_params.disc.dist = sampler_params.disc.dist
            if 'alpha' in sampler_params:
                ce_params.alpha = sampler_params.alpha
            if 'thres' in sampler_params:
                ce_params.thres = sampler_params.thres
        sampler = FeatureSampler.crossEntropySamplerFor(
            sample_space, ce_params=ce_params)
        return 'ce', sampler
    if sampler_type == 'mab':
        if sampler_params is None:
            mab_params = default_sampler_params('mab')
        else:
            mab_params = default_sampler_params('mab')
            if 'cont' in sampler_params:
                if 'buckets' in sampler_params.cont:
                    mab_params.cont.buckets = sampler_params.cont.buckets
                if 'dist' in sampler_params.cont:
                    mab_params.cont.dist = sampler_params.cont.dist
            if 'dist' in sampler_params.disc:
                mab_params.disc.dist = sampler_params.disc.dist
            if 'alpha' in sampler_params:
                mab_params.alpha = sampler_params.alpha
            if 'thres' in sampler_params:
                mab_params.thres = sampler_params.thres
            if 'priority_graph' in sampler_params:
                mab_params.priority_graph = sampler_params.priority_graph
        sampler = FeatureSampler.multiArmedBanditSamplerFor(
            sample_space, mab_params=mab_params)
        return 'mab', sampler
    if sampler_type == 'eg':
        if sampler_params is None:
            eg_params = default_sampler_params('eg')
        else:
            eg_params = default_sampler_params('eg')
            if 'cont' in sampler_params:
                if 'buckets' in sampler_params.cont:
                    eg_params.cont.buckets = sampler_params.cont.buckets
                if 'dist' in sampler_params.cont:
                    eg_params.cont.dist = sampler_params.cont.dist
            if 'dist' in sampler_params.disc:
                eg_params.disc.dist = sampler_params.disc.dist
            if 'alpha' in sampler_params:
                eg_params.alpha = sampler_params.alpha
            if 'thres' in sampler_params:
                eg_params.thres = sampler_params.thres
        sampler = FeatureSampler.epsilonGreedySamplerFor(
            sample_space, eg_params=eg_params)
        return 'eg', sampler

    if sampler_type == 'bo':
        if sampler_params is None:
            bo_params = default_sampler_params('bo')
        else:
            bo_params = default_sampler_params('bo')
            bo_params.update(sampler_params)
        sampler = FeatureSampler.bayesianOptimizationSamplerFor(
            sample_space, BO_params=bo_params)
        return 'bo', sampler

    raise ValueError(f'unknown sampler type "{sampler_type}"')

class Server:
    """Generic server for communicating with an external simulator."""
    def __init__(self, sampling_data, monitor, options={}):
        defaults = DotMap(port=8888, bufsize=4096, maxreqs=5)
        defaults.update(options)
        self.monitor = monitor
        self.lastValue = None
        self.port = defaults.port
        self.bufsize = defaults.bufsize
        self.maxreqs = defaults.maxreqs
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.host = '127.0.0.1'
        self.socket.bind((self.host, self.port))
        self.socket.listen(self.maxreqs)

        if sampling_data.sampler is not None:
            self.sampler_type = ('random' if sampling_data.sampler_type is None
                                 else sampling_data.sampler_type)
            self.sampler = sampling_data.sampler
            self.sample_space = (self.sampler.space
                                 if sampling_data.sample_space is None
                                 else sampling_data.sample_space)

        elif sampling_data.sampler_type is None:
            feature_space = {}
            for space_name in sampling_data.sample_space:
                space = sampling_data.sample_space[space_name]
                feature_space[space_name] = Feature(space)
            self.sample_space = FeatureSpace(feature_space)
            self.sampler_type = 'random'
            self.sampler = FeatureSampler.samplerFor(self.sample_space)
            self.sample_space = self.sampler.space
        else:
            feature_space = {}
            for space_name in sampling_data.sample_space:
                space = sampling_data.sample_space[space_name]
                feature_space[space_name] = Feature(space)
            self.sample_space = FeatureSpace(feature_space)
            params = (None if 'sampler_params' not in sampling_data
                      else sampling_data.sampler_params)
            self.sampler_type, self.sampler = choose_sampler(
                sample_space=self.sample_space,
                sampler_type=sampling_data.sampler_type,
                sampler_params=params
            )


    def listen(self):
        client_socket, addr = self.socket.accept()
        self.client_socket = client_socket

    def receive(self):
        data = []
        while True:
            msg = self.client_socket.recv(self.bufsize)
            if not msg:
                break
            data.append(msg)
        simulation_data = self.decode(b"".join(data))
        return simulation_data

    def send(self, sample):
        msg = self.encode(sample)
        self.client_socket.send(msg)
        self.client_socket.shutdown(socket.SHUT_WR)

    def encode(self, sample):
        return dill.dumps(sample)

    def decode(self, data):
        return dill.loads(data)

    def terminate(self):
        self.socket.close()
        
    def close_connection(self):
        self.client_socket.close()

    def get_sample(self, feedback):
        return self.sampler.nextSample(feedback)

    def flatten_sample(self, sample):
        return self.sampler.space.flatten(sample)

    def evaluate_sample(self, sample):
        self.listen()
        self.send(sample)
        simulation_data = self.receive()
        self.close_connection()
        value = (0 if self.monitor is None
                 else self.monitor.evaluate(simulation_data))
        return value

    def run_server(self):
        start = time.time()
        sample = self.get_sample(self.lastValue)
        after_sampling = time.time()
        self.lastValue = self.evaluate_sample(sample)
        after_simulation = time.time()
        timings = ServerTimings(sample_time=(after_sampling - start),
                                simulate_time=(after_simulation - after_sampling))
        return sample, self.lastValue, timings

try:
    import ray
    @ray.remote
    class ParallelServer(Server):
        pass
except ModuleNotFoundError:
    class ParallelServer(Server):
        def __init__(*args, **kwargs):
            raise RuntimeError('ParallelServer requires ray to be installed')

@dataclass
class ServerTimings:
    sample_time: float
    simulate_time: float




import functools

class frozendict(dict):
    """An immutable dictionary. Hashable if its values are."""

    def _immutableError(self, *args, **kwargs):
        raise TypeError('frozendict is immutable')

    __setitem__ = _immutableError
    __delitem__ = _immutableError
    clear = _immutableError
    pop = _immutableError
    popitem = _immutableError
    setdefault = _immutableError
    update = _immutableError

    @functools.cached_property
    def _hash(self):
        return hash(tuple(self.items()))

    def __hash__(self):
        return self._hash



"""Simulated annealing samplers"""

from verifai.samplers.domain_sampler import BoxSampler, SamplingError
import math
import numpy as np

def proposal_func(sample, iteration, decay_rate, num_variables):

    decayed_width = decay_rate**iteration*1
    lower_bound = [(sample[i] - decayed_width/2) for i in range(num_variables)]
    upper_bound = [(sample[i] + decayed_width/2) for i in range(num_variables)]

    # check whether updated range for each parameter is within original ranges
    for index in range(num_variables):
        if lower_bound[index] < 0:
            lower_bound[index] = 0
        if upper_bound[index] > 1:
            upper_bound[index] = 1

    return np.random.uniform(lower_bound, upper_bound)


class SimulatedAnnealingSampler(BoxSampler):
    def __init__(self, domain, sa_params):
        super().__init__(domain)

        ## Initialize three functions :
        self.temp_f = sa_params.get('temp_f', lambda t: 0.8*t)
        self.iter_f = sa_params.get('iter_f',
                                    lambda length: int(math.ceil(1.1*length)))
        self.proposal_f = sa_params.get('proposal_f', proposal_func)

        ## Initialize Parameters :
        self.reset_temp = sa_params.get('reset_temp', 0.01)
        
        self.T = self.init_T = sa_params.T
        self.decay_rate = sa_params.decay_rate
        self.iterations = sa_params.iterations
        self.num_epoch = sa_params.num_epoch
        self.num_iter_in_epoch = 0
        self.old_sample = self.best_sample = None
        self.old_loss = None
        self.is_first_sample = True

    def updateVector(self, vector, info, rho):
        if rho is None:    # First sample
            assert self.old_sample is None
            self.is_first_sample = False
            return

        if self.old_loss is None:
            self.old_loss = rho
        else:
            ## Compute the loss of the sample
            new_loss = rho

            # Compute the probability of whether this sample should be accepted
            alpha = min(1, np.exp((self.old_loss - new_loss)/self.T))

            ## Note that here objective is to MINIMIZE the loss
            if ((new_loss < self.old_loss) or (np.random.uniform() < alpha)):
                # Accept proposed solution
                self.old_loss = new_loss
                self.old_sample = vector

            ### Update Parameters before Returning ###
            # update temperature
            self.T = self.temp_f(self.T)

            # update iteration count within an epoch and total iteration count
            self.num_iter_in_epoch += 1

            ### Check for Temperature cooling below a threshold or
            ### Executing given # of iterations for an epoch
            if (self.T < self.reset_temp
                or self.num_iter_in_epoch > self.iterations):
                ## the temp is below reset threshold
                self.T = self.init_T
                self.iterations = self.iter_f(self.iterations) # update
                self.num_epoch -= 1 # starting new round of epoch
                self.num_iter_in_epoch = 0

    def getVector(self):
        # num_epoch := total # of rise of temp
        if self.num_epoch <= 0:
            raise TerminationException("all epochs finished")

        ## Sample the next scenario
        if self.is_first_sample:
            old_sample = np.random.uniform(0,1, self.dimension)
            self.old_sample = old_sample
            self.old_loss = None

            ### Update Parameters before Returning ###
            # update temperature
            self.T = self.temp_f(self.T)

            # update iteration count within an epoch and total iteration count
            self.num_iter_in_epoch += 1
            return old_sample, None
        else:
            new_sample = self.proposal_f(self.old_sample, self.num_iter_in_epoch,
                                        self.decay_rate, self.dimension)
            self.last_sample = new_sample

            return new_sample, None



import numpy as np
from verifai.samplers.domain_sampler import BoxSampler, DiscreteBoxSampler, \
    DomainSampler, SplitSampler
from verifai.samplers.random_sampler import RandomSampler
from verifai.samplers.cross_entropy import DiscreteCrossEntropySampler

class EpsilonGreedySampler(DomainSampler):
    def __init__(self, domain, eg_params):
        super().__init__(domain)
        self.alpha = eg_params.alpha
        self.thres = eg_params.thres
        self.cont_buckets = eg_params.cont.buckets
        self.cont_dist = eg_params.cont.dist
        self.disc_dist = eg_params.disc.dist
        self.cont_ce = lambda domain: ContinuousEpsilonGreedySampler(domain=domain,
                                                     buckets=self.cont_buckets,
                                                     dist=self.cont_dist,
                                                     alpha=self.alpha,
                                                     thres=self.thres)
        self.disc_ce = lambda domain: DiscreteEpsilonGreedySampler(domain=domain,
                                                   dist=self.disc_dist,
                                                   alpha=self.alpha,
                                                   thres=self.thres)
        partition = (
            (lambda d: d.standardizedDimension > 0, self.cont_ce),
            (lambda d: d.standardizedIntervals, self.disc_ce)
        )
        self.split_sampler = SplitSampler.fromPartition(domain,
                                                        partition,
                                                        RandomSampler)
        self.cont_sampler, self.disc_sampler = None, None
        self.rand_sampler = None
        for subsampler in self.split_sampler.samplers:
            if isinstance(subsampler, ContinuousEpsilonGreedySampler):
                assert self.cont_sampler is None
                self.cont_sampler = subsampler
            elif isinstance(subsampler, DiscreteEpsilonGreedySampler):
                assert self.disc_sampler is None
                self.disc_sampler = subsampler
            else:
                assert isinstance(subsampler, RandomSampler)
                assert self.rand_sampler is None
                self.rand_sampler = subsampler

    def getSample(self):
        return self.split_sampler.getSample()

    def update(self, sample, info, rho):
        self.split_sampler.update(sample, info, rho)

class ContinuousEpsilonGreedySampler(BoxSampler):
    def __init__(self, domain, alpha, thres,
                 buckets=10, dist=None, epsilon=0.5):
        super().__init__(domain)
        if isinstance(buckets, int):
            buckets = np.ones(self.dimension) * buckets
        elif len(buckets) > 1:
            assert len(buckets) == self.dimension
        else:
            buckets = np.ones(self.dimension) * buckets[0]
        if dist is not None:
            assert (len(dist) == len(buckets))
        if dist is None:
            dist = np.array([np.ones(int(b))/b for b in buckets])
        self.buckets = buckets
        self.dist = dist
        self.alpha = alpha
        self.thres = thres
        self.current_sample = None
        self.counts = np.array([np.ones(int(b)) for b in buckets])
        self.errors = np.array([np.zeros(int(b)) for b in buckets])
        self.t = 1
        self.epsilon = epsilon
        self.sample_randomly = np.random.uniform() < self.epsilon

    def nextVector(self, feedback=None):
        self.update(None, self.current_sample, feedback)
        return self.generateSample()
    
    def generateSample(self):
        if self.sample_randomly:
            bucket_samples = np.array([np.random.choice(int(b))
                                    for i, b in enumerate(self.buckets)])
        else:
            bucket_samples = np.array([np.random.choice(int(b), p=self.dist[i])
                                    for i, b in enumerate(self.buckets)])
        self.current_sample = bucket_samples
        ret = tuple(np.random.uniform(bs, bs+1.)/b for b, bs
              in zip(self.buckets, bucket_samples))
        return ret, bucket_samples
    
    def updateVector(self, vector, info, rho):
        if rho is None or rho >= self.thres:
            return
        self.t += 1
        if self.t % 100 == 0:
            self.epsilon *= 0.5
        update_dist = np.array([np.zeros(int(b)) for b in self.buckets])
        for i, (ud, b) in enumerate(zip(update_dist, info)):
            ud[b] = 1.
        self.dist = self.alpha*self.dist + (1-self.alpha)*update_dist
        self.sample_randomly = np.random.uniform() < self.epsilon
        # print(self.errors / self.counts)

class DiscreteEpsilonGreedySampler(DiscreteCrossEntropySampler):
    pass